from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING
from typing import Awaitable
from typing import Callable

from litestar.exceptions import ImproperlyConfiguredException

if TYPE_CHECKING:
    from svcs import Registry


@dataclass(frozen=True)
class SvcsPluginConfig:
    """The config for the `SvcsPlugin`."""

    registry: Registry | None = None
    """The `svcs.Registry` instance.

    If provided, this is used and if not a Registry is created with `registry_factory`.
    """

    registry_factory: (
        Callable[[], Registry] | Callable[[], Awaitable[Registry]] | None
    ) = None
    """A factory function to get registry.

    This is called on app startup if the registry instance has not been
    provided. This can be async or sync.
    """

    container_dependency_key: str = "svcs_container"
    """The dependency key to use for injecting `svcs.Container`."""

    registry_state_key: str = "svcs_registry"
    """The key used to store the Registry within the Litestar `State` instance."""

    def __post_init__(self) -> None:
        if self.registry is None and self.registry_factory is None:
            msg = "either `registry` or `registry_factory` must be provided"
            raise ImproperlyConfiguredException(
                msg,
            )

        if self.registry is not None and self.registry_factory is not None:
            msg = "only one of `registry` and `registry_factory` must be provided"
            raise ImproperlyConfiguredException(
                msg,
            )



from __future__ import annotations

from contextlib import asynccontextmanager
from inspect import isawaitable
from typing import TYPE_CHECKING
from typing import AsyncGenerator
from typing import Final

from litestar.di import Provide
from litestar.exceptions import ImproperlyConfiguredException
from litestar.params import Dependency
from litestar.plugins import InitPluginProtocol
from svcs import Container
from svcs import Registry
from typing_extensions import override

if TYPE_CHECKING:
    from litestar.app import Litestar
    from litestar.config.app import AppConfig
    from litestar.datastructures.state import State

    from litestar_svcs.config import SvcsPluginConfig


class SvcsPlugin(InitPluginProtocol):
    """The plugin for integrating `svcs` to a `Litestar` application."""

    _registry: Registry
    _REGISTRY_DEPENDENCY_NAME: Final[str] = "svcs_registry"

    def __init__(self, config: SvcsPluginConfig) -> None:
        self._config = config

    @property
    def registry(self) -> Registry:
        """Get the registry instance.

        This will raise a `RuntimeError` if the application has not started yet
        and the registry has not been setup yet.
        """

        try:
            return self._registry
        except AttributeError as ex:
            msg = "the svcs registry has not been setup yet"
            raise RuntimeError(msg) from ex

    @asynccontextmanager
    async def get_container(self) -> AsyncGenerator[Container, None]:
        """Get a `svcs.Container` instance.

        This can be used in places where dependency injection is not possible
        within `litestar` such as in middlewares.
        """

        async with Container(self.registry) as container:
            yield container

    def provide_registry(self, state: State) -> Registry:
        """Provide the registry given the Litestar state.

        Args:
        ----
            state: The `Litestar` state instance.
        """

        registry = state.get(self._config.registry_state_key)

        assert registry is not None

        return registry

    async def provide_container(
        self,
        svcs_registry: Registry = Dependency(skip_validation=True),  # noqa: B008
    ) -> AsyncGenerator[Container, None]:
        """Provide the container from the given registry.

        Args:
        ----
            svcs_registry: A `svcs.Registry` instance.
        """

        async with Container(svcs_registry) as container:
            yield container

    @asynccontextmanager
    async def lifespan(self, app: Litestar) -> AsyncGenerator[None, None]:
        """Add the lifespan to the application.

        This is where the registry is created and injected into the app state.
        """
        registry: Registry
        if self._config.registry:
            registry = self._config.registry
        elif self._config.registry_factory:
            _registry = self._config.registry_factory()
            if isawaitable(_registry):
                self._registry = registry = await _registry
            else:
                self._registry = registry = _registry  # type: ignore[assignment]
        else:
            msg = "either `registry` or `registry_factory` must be provided"
            raise ImproperlyConfiguredException(
                msg,
            )
        app.state.update({self._config.registry_state_key: registry})

        try:
            yield
        finally:
            await registry.aclose()

    @override
    def on_app_init(self, app_config: AppConfig) -> AppConfig:
        app_config.dependencies.update(
            {
                self._config.container_dependency_key: Provide(self.provide_container),
                self._REGISTRY_DEPENDENCY_NAME: Provide(
                    self.provide_registry,
                    use_cache=True,
                    sync_to_thread=False,
                ),
            },
        )
        app_config.lifespan.append(self.lifespan)
        app_config.signature_namespace.update(
            {
                "Container": Container,
                "Registry": Registry,
            },
        )

        return app_config



from __future__ import annotations

import pytest
from litestar.exceptions import ImproperlyConfiguredException
from svcs import Registry

from litestar_svcs.config import SvcsPluginConfig


def test_no_registry_given() -> None:
    """Instantiation should raise an error if neither `registry` nor `registry_factory` is provided."""

    with pytest.raises(
        ImproperlyConfiguredException,
        match="either `registry` or `registry_factory` must be provided",
    ):
        _ = SvcsPluginConfig()


def test_both_registry_and_factory_given() -> None:
    """Instantiation should raise an error if both `registry and `registry_facotry` is provided."""

    registry = Registry()

    def registry_factory() -> Registry:
        return registry

    with pytest.raises(
        ImproperlyConfiguredException,
        match="only one of `registry` and `registry_factory` must be provided",
    ):
        _ = SvcsPluginConfig(registry=registry, registry_factory=registry_factory)



from __future__ import annotations

from http import HTTPStatus
from typing import Callable

import pytest
from litestar import get
from litestar.params import Dependency
from litestar.testing import create_test_client
from svcs import Container
from svcs import Registry

from litestar_svcs.config import SvcsPluginConfig
from litestar_svcs.plugin import SvcsPlugin

NUMBER = 10
STRING = "foo"


def _registry_factory() -> Registry:
    registry = Registry()

    registry.register_factory(int, lambda: NUMBER)
    registry.register_factory(str, lambda: STRING)

    return registry


async def _async_registry_factory() -> Registry:
    return _registry_factory()


@pytest.mark.parametrize(
    "registry",
    (_registry_factory(), _registry_factory, _async_registry_factory),
)
def test_container_injected(registry: Registry | Callable[[], Registry]) -> None:
    """Ensure the container is injected as a dependency."""

    @get("/", sync_to_thread=False)
    def get_handler(
        svcs_container: Container = Dependency(skip_validation=True),
    ) -> None:
        assert svcs_container.get(int) == NUMBER
        assert svcs_container.get(str) == STRING

    config = (
        SvcsPluginConfig(registry)
        if isinstance(registry, Registry)
        else SvcsPluginConfig(registry_factory=registry)
    )
    plugin = SvcsPlugin(config)

    with create_test_client([get_handler], plugins=[plugin]) as client:
        response = client.get("/")

        assert response.status_code == HTTPStatus.OK


def test_container_injected_custom_key() -> None:
    """Test container is injected when using a custom dependency key."""

    @get("/", sync_to_thread=False)
    def get_handler(
        container: Container = Dependency(skip_validation=True),
    ) -> None:
        assert container.get(int) == NUMBER
        assert container.get(str) == STRING

    registry = _registry_factory()
    config = SvcsPluginConfig(registry, container_dependency_key="container")
    plugin = SvcsPlugin(config)

    with create_test_client([get_handler], plugins=[plugin]) as client:
        response = client.get("/")

        assert response.status_code == HTTPStatus.OK


@pytest.mark.parametrize(
    "registry",
    (_registry_factory(), _registry_factory, _async_registry_factory),
)
def test_registry_injected(registry: Registry | Callable[[], Registry]) -> None:
    """Ensure the registry is injected as a dependency."""

    @get("/", sync_to_thread=False)
    def get_handler(svcs_registry: Registry = Dependency(skip_validation=True)) -> None:
        container = Container(svcs_registry)

        assert container.get(int) == NUMBER
        assert container.get(str) == STRING

    config = (
        SvcsPluginConfig(registry)
        if isinstance(registry, Registry)
        else SvcsPluginConfig(registry_factory=registry)
    )
    plugin = SvcsPlugin(config)

    with create_test_client([get_handler], plugins=[plugin]) as client:
        response = client.get("/")

        assert response.status_code == HTTPStatus.OK


def test_registry_closed() -> None:
    """The registry must be closed after the app is shutdown."""

    @get("/", sync_to_thread=False)
    def get_handler() -> None:
        ...

    registry = _registry_factory()
    config = SvcsPluginConfig(registry)
    plugin = SvcsPlugin(config)

    with create_test_client([get_handler], plugins=[plugin]) as client:
        response = client.get("/")

        assert response.status_code == HTTPStatus.OK

    # The registry clears all services when it's closed.
    assert int not in registry
    assert str not in registry
